#include <windows.h>
#include <winuser.h>
#include <string.h>
#include <malloc.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

#include "launch.h"

#if defined(__CYGWIN__)
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/cygwin.h>
 #include <sys/unistd.h>
#else
 #include <direct.h>
#endif


char buffer[1024];

/* Copy cygwin environment variables to the Windows environment if they're not
 * already there. */
static void setup_win_environ(void)
{
    char **envp = environ;
    char *var, *val;
    char curval[2];

    while (envp && *envp) {
        var = strdup(*envp++);
        val = strchr(var, '=');
		if (val != NULL)
		{
			*val++ = '\0';
        
			if (GetEnvironmentVariable(var, curval, 2) == 0 && 
			    GetLastError() == ERROR_ENVVAR_NOT_FOUND) 
			{
				SetEnvironmentVariable(var, val);
			}
		}
        free(var);
    }
}
BOOL setup_invisible_console()
{
   HWINSTA h, horig;
   USEROBJECTFLAGS oi;
   DWORD len;
   BOOL b = FALSE; 
   HMODULE lib = NULL;
   HWINSTA WINAPI (*GetProcessWindowStationFP)(void) = NULL;
   BOOL WINAPI (*GetUserObjectInformationFP)(HANDLE, int, PVOID, DWORD, PDWORD) = NULL;
   HWINSTA WINAPI (*CreateWindowStationFP)(LPCWSTR, DWORD, DWORD, LPSECURITY_ATTRIBUTES) = NULL;
   BOOL WINAPI (*SetProcessWindowStationFP)(HWINSTA) = NULL;
   BOOL WINAPI (*CloseWindowStationFP)(HWINSTA) = NULL;

   /* until we have a mechanism of determining whether a given HANDLE 
    * returned by GetStdHandles actually derives from a console, 
    * unconditionally call FreeConsole() on all OSes under all conditions.
    * See comments in configure_startupinfo(). 
    */
   FreeConsole();

   /* First, set up function pointers */
   if (lib = LoadLibrary ("user32.dll"))
   {
       GetProcessWindowStationFP = (HWINSTA WINAPI (*)(void))
           GetProcAddress (lib, "GetProcessWindowStation");
       GetUserObjectInformationFP = (BOOL WINAPI (*)(HANDLE, int, PVOID, DWORD, PDWORD))
           GetProcAddress (lib, "GetUserObjectInformationW"); /* ugly! */
       CreateWindowStationFP = (HWINSTA WINAPI (*)(LPCWSTR, DWORD, DWORD, LPSECURITY_ATTRIBUTES))
           GetProcAddress (lib, "CreateWindowStationW"); /* ugly */
       SetProcessWindowStationFP = (BOOL WINAPI (*)(HWINSTA))
           GetProcAddress (lib, "SetProcessWindowStation");
       CloseWindowStationFP = (BOOL WINAPI (*)(HWINSTA))
           GetProcAddress (lib, "CloseWindowStation");

       if (GetProcessWindowStationFP && 
            GetUserObjectInformationFP &&
            CreateWindowStationFP &&
            SetProcessWindowStationFP &&
            CloseWindowStationFP)
       {
           /* Then, do the work */
           FreeConsole();
           h = horig = (*GetProcessWindowStationFP)();
           if (!horig 
               || !(*GetUserObjectInformationFP) (horig, UOI_FLAGS, &oi, sizeof (oi), &len )
               || !(oi.dwFlags & WSF_VISIBLE))
           {
               b = AllocConsole();
           }
           else
           {
               h = (*CreateWindowStationFP) (NULL, 0, STANDARD_RIGHTS_READ, NULL);
               if (h)
               {
                   b = (*SetProcessWindowStationFP) (h);
               }
               b = AllocConsole();
               if (horig && h && h != horig && (*SetProcessWindowStationFP) (horig))
               {
                   (*CloseWindowStationFP) (h);
               }
           }
           return b;
       }
   }
   /* otherwise, fail */ 
   return FALSE;
}

/* returns FALSE only on error conditions (not impl) */
BOOL configure_startupinfo(STARTUPINFO* psi, BOOL bHaveInvisConsole,
                           BOOL *bUsingPipes,
                           HANDLE* hpToChild,  HANDLE* hpFromChild,
                           HANDLE* hpToParent, HANDLE* hpFromParent)
{
    CONSOLE_CURSOR_INFO cci;
    SECURITY_ATTRIBUTES handle_attrs;
    HANDLE hpStdInput[2];
    HANDLE hpStdOutput[2];

    ZeroMemory (psi, sizeof (STARTUPINFO));
    psi->cb = sizeof (STARTUPINFO);
    psi->hStdInput   = GetStdHandle(STD_INPUT_HANDLE);
    psi->hStdOutput  = GetStdHandle(STD_OUTPUT_HANDLE);
    psi->hStdError   = GetStdHandle(STD_ERROR_HANDLE);
    psi->dwFlags     = STARTF_USESHOWWINDOW | STARTF_USESTDHANDLES;
    psi->wShowWindow = SW_HIDE;

    /* foo() is some magic mechanism for determining that the HANDLEs 
     * returned by GetStdHandle() are from a console, and not redirected
     * or ptys of some sort.  If we have such a mechanism, then the 
     * unconditional FreeConsole() at the top of setup_invisible_console()
     * should be removed.
     */
/*
    if (foo())
    {
       *bUsingPipes = FALSE;
       return TRUE;
    }
*/

    /* but for now, the only way we KNOW we have a console is
     * if we created it ourselves
     */
    if (bHaveInvisConsole)
    {
       *bUsingPipes = FALSE;
       return TRUE;
    }

    /* otherwise, set up pipes */
    *bUsingPipes = TRUE;

    handle_attrs.nLength = sizeof(SECURITY_ATTRIBUTES);
    handle_attrs.bInheritHandle = TRUE;
    handle_attrs.lpSecurityDescriptor = NULL;
 
    /* create a pipe for child's stdin.  Don't allow child to */
    /* inherit the write end of the pipe.                     */
    CreatePipe (&hpStdInput[0], &hpStdInput[1], &handle_attrs, 0);
    SetHandleInformation ( hpStdInput[1], HANDLE_FLAG_INHERIT, 0);
 
    /* create a pipe for child's stdout.  Don't allow child to */
    /* inherit the read end of the pipe.                       */
    CreatePipe (&hpStdOutput[0], &hpStdOutput[1], &handle_attrs, 0);
    SetHandleInformation ( hpStdOutput[0], HANDLE_FLAG_INHERIT, 0);

    psi->hStdInput   = hpStdInput[0];
    psi->hStdOutput  = hpStdOutput[1];
    psi->hStdError   = hpStdOutput[1];

    *hpToChild   = hpStdInput[1];
    *hpFromChild = hpStdOutput[0];
    *hpToParent  = hpStdOutput[1];
    *hpFromParent= hpStdInput[0];

    return TRUE;
}
int start_child(char* cmdline, int wait_for_child)
{
   STARTUPINFO start;
   PROCESS_INFORMATION child;
   DWORD retval;
   BOOL bFuncReturn;
   BOOL bHaveInvisConsole;
   BOOL bUsingPipes;
   HANDLE hToChild, hFromChild;
   HANDLE hToParent, hFromParent;

   setup_win_environ();

#ifdef DEBUG_FORCE_PIPES
   bHaveInvisConsole = FALSE;
   FreeConsole();
#else
   bHaveInvisConsole = setup_invisible_console();
#endif

   if (!configure_startupinfo(&start, bHaveInvisConsole,
                              &bUsingPipes,
                              &hToChild, &hFromChild, 
                              &hToParent, &hFromParent))
      error("could not start %s",cmdline);

#ifdef DEBUG
   message("Using Pipes: %s", (bUsingPipes ? "yes" : "no"));
#endif

   ZeroMemory( &child, sizeof(PROCESS_INFORMATION) );

   bFuncReturn = CreateProcess (NULL,
       cmdline, /* command line                        */
       NULL,    /* process security attributes         */
       NULL,    /* primary thread security attributes  */
       TRUE,    /* handles are inherited,              */
       0,       /* creation flags                      */
       NULL,    /* use parent's environment            */
       NULL,    /* use parent's current directory      */
       &start,  /* STARTUPINFO pointer                 */
       &child); /* receives PROCESS_INFORMATION        */

   if (bUsingPipes)
   {
       CloseHandle(hFromParent);
       CloseHandle(hToParent);
   }

   if (bFuncReturn)
   {
      if (wait_for_child)
      {
          if (bUsingPipes)
          {
              HANDLE handles[2] = { child.hProcess, hFromChild };
              COMMTIMEOUTS timeouts;
    
              /* only read bytes that are ready, and return immediately */
              GetCommTimeouts(hFromChild, &timeouts);
              timeouts.ReadIntervalTimeout = MAXDWORD;
              timeouts.ReadTotalTimeoutMultiplier = 0;
              timeouts.ReadTotalTimeoutConstant = 0;
              SetCommTimeouts(hFromChild, &timeouts);
    
              while (WaitForMultipleObjects (2, handles, FALSE, INFINITE) == WAIT_OBJECT_0 + 1)
              {
                  char buffer[1024];
                  DWORD dwRead;
                  ReadFile (hFromChild, buffer, 1024, &dwRead, NULL);
              }
          }
          else
          {
              WaitForSingleObject (child.hProcess, INFINITE);
          }
          GetExitCodeProcess (child.hProcess, &retval);
      } 
      CloseHandle (child.hThread);
      CloseHandle (child.hProcess);
      if (bUsingPipes)
      {
          CloseHandle (hFromChild);
          CloseHandle (hToChild);
      }
   }
   else
   {
      if (bUsingPipes)
      {
          CloseHandle (hFromChild);
          CloseHandle (hToChild);
      }
      error("could not start %s",cmdline);
   }
   return retval;
}

int build_cmdline(char* new_cmdline, char* exec, int argc, char* argv[])
{
   int retval = FALSE;
   int first_arg = 1;
   int i;
   int char_cnt = 0;
   /*
    * look for "-wait" as first true argument; we'll apply that ourselves
    */
   if ((argc >= 2) && (stricmp(argv[1],"-wait") == 0))
   {
      retval = TRUE;
      first_arg++;
   }

   char_cnt = strlen(exec);
   for (i = first_arg; i < argc; i++)
      char_cnt += strlen(argv[i]);
   if (char_cnt > MAX_ARGS*MAX_PATH) /* then we ran out of room */
      error("command line too long -\n%s",new_cmdline);
   
   strcpy(new_cmdline,exec);
   for (i = first_arg; i < argc; i++)
   {
      strcat(new_cmdline," ");
      strcat(new_cmdline,argv[i]);
   }
   return retval;
}
/* process exec_arg : if it
 * NATIVE:
 *  1) starts with '\\' or '/', it's a root-path and leave it alone
 *  2) starts with 'x:\\' or 'x:/', it's a root-path and leave it alone
 *  3) starts with '.\\' or './', two possible meanings:
 *       1) exec is in the current directory
 *       2) exec in same directory as this program
 *  4) otherwise, search path (and _prepend_ "." to the path!!!)
 *  5) convert all '/' to '\\'
 * CYGWIN
 *  1) starts with '\\' or '/', it's a root-path and leave it alone
 *  2) starts with 'x:\\' or 'x:/', it's a root-path and leave it alone
 *  3) starts with '.\\' or './', two possible meanings:
 *       1) exec is in the current directory
 *       2) exec in same directory as this program
 *  4) otherwise, search path (and _prepend_ "." to the path!!!)
 *  5) convert to cygwin-style path to resolve symlinks within the pathspec
 *  6) check filename: if it's a symlink, resolve it by peeking inside
 *  7) convert to win32-style path+filename since we're using Windows 
 *       createProcess() to launch
 */
void process_execname(char *exec, const char* execname,const char* execpath )
{
   char* orig_pathlist;
   char* pathlist;
   char exec_tmp[MAX_PATH + FILENAME_MAX + 100];
   char exec_tmp2[MAX_PATH + FILENAME_MAX + 100];
   char buf[MAX_PATH + FILENAME_MAX + 100];
   int i,j;

   int len = 0;
   /* 
    * STARTS WITH / or \ 
    * execpath NOT used
    */
   if ((execname[0] == '\\') || (execname[0] == '/'))
   {
#if defined(__CYGWIN__)
      strcpy(exec_tmp,execname);
#else    
      exec_tmp[0] = ((char) (_getdrive() + ((int) 'A') - 1));
      exec_tmp[1] = ':';
      exec_tmp[2] = '\0';
      strcat(exec_tmp,execname);
#endif
      Trace(("/ -\nexec_tmp\t%s\nexecname\t%s\nexecpath\t%s\n",
             exec_tmp,execname,execpath));
      if (! fileExistsMulti(exec_tmp2,NULL,exec_tmp,exts,NUM_EXTENSIONS) )
      {
          j = 0;
          for (i = 0; i < NUM_EXTENSIONS; i++)
              j += sprintf(buf + j," [%d]: %s\n",i+1,exts[i]);
          error("Couldn't locate %s\nI tried appending the following "
                "extensions: \n%s",exec_tmp,buf);
      }
      Trace((exec_tmp2));
   }
   /* 
    * STARTS WITH x:\ or x:/
    * execpath NOT used
    */
    else if ((strlen(execname) > 3) && /* avoid boundary errors */
       (execname[1] == ':') &&
       ((execname[2] == '\\') || (execname[2] == '/')))
   {
      strcpy(exec_tmp,execname);       
      Trace(("x: -\nexec_tmp\t%s\nexecname\t%s\nexecpath\t%s\n",
             exec_tmp,execname,execpath));
      if (! fileExistsMulti(exec_tmp2,NULL,exec_tmp,exts,NUM_EXTENSIONS) )
      {
          j = 0;
          for (i = 0; i < NUM_EXTENSIONS; i++)
              j += sprintf(buf + j," [%d]: %s\n",i+1,exts[i]);
          error("Couldn't locate %s\nI tried appending the following "
                "extensions: \n%s",exec_tmp,buf);
      }
      Trace((exec_tmp2));
   }
   /* 
    * STARTS WITH ./ or .\
    */
   else if ((execname[0] == '.') &&
            ((execname[1] == '\\') || (execname[1] == '/')))
   {
      if (((char*) getcwd(exec_tmp,MAX_PATH))==NULL)
         error("can't find current working directory");
      if (! fileExistsMulti(exec_tmp2,exec_tmp,&(execname[2]),
                            exts,NUM_EXTENSIONS) )
          if (! fileExistsMulti(exec_tmp2,execpath,&(execname[2]),
                                exts,NUM_EXTENSIONS) )
          {
              j = 0;
              for (i = 0; i < NUM_EXTENSIONS; i++)
                  j += sprintf(buf + j," [%d]: %s\n",i+1,exts[i]);
              error("Couldn't locate %s\n"
                    "I looked in the following directories:\n [1]: %s\n [2]: %s\n"
                    "I also tried appending the following "
                    "extensions: \n%s",execname,exec_tmp,execpath,buf);
          }
      Trace((exec_tmp2));
   }
   /*
    * OTHERWISE, SEARCH PATH (prepend '.' and run.exe's directory)
    * can't use fileExistsMulti because we want to search entire path
    * for exts[0], then for exts[1], etc.
    */
   else
   {
      orig_pathlist = getenv("PATH");
      if ((pathlist = malloc (strlen(orig_pathlist)
                              + strlen(".") 
                              + strlen(execpath)+ 3)) == NULL)
         error("internal error - out of memory");
      strcpy(pathlist,".");
      strcat(pathlist,SEP_CHARS);
      strcat(pathlist,execpath);
      strcat(pathlist,SEP_CHARS);
      strcat(pathlist,orig_pathlist);

      Trace((pathlist));
      for (i = 0; i < NUM_EXTENSIONS; i++)
      {
          strcpy(exec_tmp,execname);
          strcat(exec_tmp,exts[i]);
          pfopen(exec_tmp2,exec_tmp,pathlist);
          if (fileExists(NULL,NULL,exec_tmp2))
              break;
          exec_tmp2[0] = '\0';
      }
      Trace(("exec_tmp\t%s\npathlist\t%s\n",exec_tmp2,pathlist));

      free(pathlist);
      if (exec_tmp2[0] == '\0')
      {
          j = 0;
          for (i = 0; i < NUM_EXTENSIONS; i++)
              j += sprintf(buf + j," [%d]: %s\n",i+1,exts[i]);
          error("Couldn't find %s anywhere.\n"
                "I even looked in the PATH \n"
                "I also tried appending the following "
                "extensions: \n%s",execname,buf);
      }
   }
/*
 * At this point, we know that exec_tmp2 contains a filename
 * and we know that exec_tmp2 exists.
 */
#if defined(__CYGWIN__)
   {
      struct stat stbuf;
      char sym_link_name[MAX_PATH+1];
      char real_name[MAX_PATH+1];
      char dummy[MAX_PATH+1];

      strcpy(exec_tmp,exec_tmp2);

      CYGWIN_CONV_TO_POSIX_PATH((exec_tmp,sym_link_name));
      Trace((sym_link_name));
      
      if (lstat(sym_link_name, &stbuf) == 0)
      {
         if ((stbuf.st_mode & S_IFLNK) == S_IFLNK)
         {
	    if (readlink(sym_link_name, real_name, sizeof(real_name)) == -1)
               error("problem reading symbolic link for %s",exec_tmp);
            else
            {
                /* if realname starts with '/' it's a rootpath */
                if (real_name[0] == '/')
                    strcpy(exec_tmp2,real_name);
                else /* otherwise, it's relative to the symlink's location */
                {
                   CYGWIN_SPLIT_PATH((sym_link_name,exec_tmp2,dummy));
                   if (!endsWith(exec_tmp2,PATH_SEP_CHAR_STR))
                      strcat(exec_tmp2,PATH_SEP_CHAR_STR);
                   strcat(exec_tmp2,real_name);
                }
            }
         }
         else /* NOT a symlink */
            strcpy(exec_tmp2, sym_link_name);
      }
      else
         error("can't locate executable - %s",sym_link_name);
   }
   CYGWIN_CONV_TO_FULL_WIN32_PATH((exec_tmp2,exec));
#else					
   strcpy (exec, exec_tmp2);
#endif  
}
int endsWith(const char* s1, const char* s2)
{
    int len1;
    int len2;
    int retval = FALSE;
    len1 = strlen(s1);
    len2 = strlen(s2);
    if (len1 - len2 >= 0)
        if (stricmp(&(s1[len1-len2]),s2) == 0)
            retval = TRUE;
    return retval;
}void strip_exe(char* s)
{
   if ((strlen(s) > 4) && /* long enough to have .exe extension */
       /* second part not evaluated (short circuit) if exec_arg too short */
       (stricmp(&(s[strlen(s)-4]),".exe") == 0))
      s[strlen(s)-4] = '\0';
}
/*
 * Uses system info to determine the path used to invoke run
 * Also attempts to deduce the target execname if "compact_invocation"
 * method was used.
 *
 * returns TRUE if compact_invocation method was used
 *   (and target execname was deduced successfully)
 * otherwise returns FALSE, and execname == run or run.exe
 */
int get_exec_name_and_path(char* execname, char* execpath)
{
   char modname[MAX_PATH];
   char* tmp_execname;
   char* p;
   int retval = FALSE;

   if (!GetModuleFileName (NULL, modname, MAX_PATH))
      error("internal error - can't find my own name");
   if ((p = strrchr (modname, '\\')) == NULL)
      error("internal error - my own name has no path\n%s",modname);
   tmp_execname = p + 1;
   p[0] = '\0';
   /* if invoked by a name like "runxemacs" then strip off
    * the "run" and let "xemacs" be execname.
    * To check for this, make that:
    *   1) first three chars are "run"
    *   2) but the string doesn't end there, or start ".exe"
    * Also, set "compact_invocation" TRUE
    */
   if ( ((tmp_execname[0] == 'r') || (tmp_execname[0] == 'R')) &&
        ((tmp_execname[1] == 'u') || (tmp_execname[1] == 'U')) &&
        ((tmp_execname[2] == 'n') || (tmp_execname[2] == 'N')) &&
        ((tmp_execname[3] != '.') && (tmp_execname[3] != '\0')) )
   {
      tmp_execname += 3;
      retval = TRUE;
   }
   else
      tmp_execname = NULL;

   if (tmp_execname == NULL)
      strcpy(execname,"");
   else
      strcpy(execname,tmp_execname);
#if defined(__CYGWIN__)
   CYGWIN_CONV_TO_POSIX_PATH((modname,execpath));
#else
   strcpy(execpath,modname);
#endif
   return retval;
}
/*
 * works like strtok, but:
 * double quotes (") suspends tokenizing until closing " reached
 * CYGWIN ONLY:
 *   additionally, backslash escapes next character, even if that
 *   next character is a delimiter. Or a double quote.
 *   WARNING: this means that backslash may NOT be a delimiter 
 */
char* my_strtok(char* s, const char* delim, char** lasts)
{
   char *spanp;
   int c, sc;
   char *tok;
   
   if ((s == NULL) && ((s = *lasts) == NULL))
      return NULL;
   /* Skip leading delimiters */
cont:
   c = *s++;
   for (spanp = (char *)delim; (sc = *spanp++) != 0;) {
      if (c == sc)
         goto cont;
   }
   if (c == 0) {		/* no non-delimiter characters */
      *lasts = NULL;
      return (NULL);
   }
   tok = s - 1;
   /*
    * Scan token (scan for delimiters: s += strcspn(s, delim), sort of).
    * Note that delim must have one NUL; we stop if we see that, too.
    * If we see a double quote, continue until next double quote, then
    *   start scanning for delimiters again.
    * CYGWIN ONLY: if we see a backslash, just copy next character -
    *   don't consider it as a delimiter even if it is in delim string.
    */
   for (;;) {
      /* if this c is ", then scan until we find next " */
      if (c == '\"')
         while ((c = *s++) != '\"')
            if (c == 0) /* oops, forgot to close the ", clean up & return */
            {
               s = NULL;
               *lasts = s;
               return (tok);
            }
#if defined(__CYGWIN__)
      if (c == '\\')
      {
         c = *s++; /* skip the backslash */
         if (c == 0) /* if escaped character is end-of-string, clean up & return */
         {
            s = NULL;
            *lasts = s;
            return (tok);
         }
         c = *s++; /* otherwise, skip the escaped character */
      }
#endif   
      spanp = (char *)delim;
      do {
         if ((sc = *spanp++) == c) {
            if (c == 0)
               s = NULL;
            else
               s[-1] = 0;
            *lasts = s;
            return (tok);
         }
      } while (sc != 0);
      c = *s++;
   }
   /* NOTREACHED */
}

#ifdef __CYGWIN__
void addPathCygwin(const char *str) 
{
    char wstr[MAX_PATH];
    DWORD envsize;
    size_t slen;
    char *buffer;

    CYGWIN_CONV_TO_FULL_WIN32_PATH((str,wstr));
    slen = strlen(wstr);

    envsize = GetEnvironmentVariable("PATH", NULL, 0);
    if (envsize == 0)
    {
        DWORD lasterror = GetLastError();
        if (lasterror == ERROR_ENVVAR_NOT_FOUND)
        {
            if (!SetEnvironmentVariable("PATH", wstr))
                error("SetEnvironmentVariable failed: 0x%x", GetLastError());
        }
        error("GetEnvironmentVariable failed: 0x%x", lasterror);
    }
    buffer = malloc(envsize + slen + 2);
    if (buffer == NULL)
        error("internal error - out of memory");

    if (!GetEnvironmentVariable("PATH", buffer, envsize))
        error("GetEnvironmentVariable failed: 0x%x", GetLastError());

    strcat(buffer, ";");
    strcat(buffer, wstr);

    if (!SetEnvironmentVariable("PATH", buffer))
        error("SetEnvironmentVariable failed: 0x%x", GetLastError());
    free(buffer);
}

#endif
void addPath(const char *str)
{
    const char *path = getenv("PATH");
    if (path == NULL)
    {
        setenv("PATH", str, 0);
    } else 
    {
        char *buffer = malloc(strlen(path) + strlen(str) + 2);
        strcpy(buffer, path);
        strcat(buffer, ":");
        strcat(buffer, str);
        setenv("PATH" , buffer, 1);
        free(buffer);
    }
#ifdef __CYGWIN__
    addPathCygwin(str);
#endif
}


int parse_cmdline_to_arg_array(char* argv[MAX_ARGS], char* cmdline)
{
   char seps[] = " \t\n";
   char* token;
   int argc = 0;
   char* lasts;

   token = my_strtok(cmdline, seps, &lasts);
   while ((token != NULL) && (argc < MAX_ARGS))
   {
       if (strcmp(token,"-p") == 0) 
       {
           token = my_strtok(NULL,seps,&lasts);
           if (token == NULL)
           {
               error("-p expects parameter");
           }

           if (*token == '"')
           {
               int l;
               token++;
               l = strlen(token);
               if (l > 0 && token[l-1] == '"')
                   token[l-1] = 0;
           }
           addPath(token);
       } else 
       {
           if ((argv[argc] = malloc(strlen(token)+1)) == NULL)
           {
               error("internal error - out of memory");
           }
           strcpy(argv[argc++],token);
       }
       token = my_strtok(NULL,seps,&lasts);
   }
   if (argc >= MAX_ARGS)
      error("too many arguments on commandline\n%s",cmdline);
   return argc;
}
int fileExistsMulti(char* fullname, const char* path, 
                    const char* name_noext, const char* exts[],
                    const int extcnt)
{
    char tryName[MAX_PATH + FILENAME_MAX];
    int i = 0;
    int retval = FALSE;
    fullname[0] = '\0';
    for (i = 0; i < extcnt; i++)
    {
        strcpy(tryName,name_noext);
        strcat(tryName,exts[i]);
        if (fileExists(fullname, path, tryName) == TRUE)
        {
            retval = TRUE;
            break;
        }
        fullname[0] = '\0';
    }
    return retval;
}
