/*
 * Copyright (c) 2006 Charles S. Wilson
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included 
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR 
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
 * OTHER DEALINGS IN THE SOFTWARE.
 */
#ifndef UTIL_H
#define UTIL_H

#define PATH_SEP_CHAR_STR "/"
#define SEP_CHAR ":"
#define SEP_CHARS ":"

char* pfopen(char *retval, const char *name, const char *dirs);
int fileExists(char* fullname, const char* path, const char* name);
void message_(int level, char* fmt, ...);
void debug_(char* fmt, ...);
int countargv(char** vector);


/* modes */
int gui_is_allowed(void);
int tty_is_allowed(void);
int mode_is_silent(void);
int mode_is_debug(void);
void set_gui_mode(int mode); /* bool: true=enabled */
void set_tty_mode(int mode); /* bool: true=enabled */
void set_silent_mode(int mode); /* bool: true=silent */
void set_debug_mode(int mode);  /* bool: true=debug on */


#if defined(DEBUG)
# define debugMsg(fmt, ...) debug_(fmt, ## __VA_ARGS__)
#else
# define debugMsg(fmt, ...) do { if (mode_is_debug()) { debug_(fmt, ## __VA_ARGS__); } } while(0)
#endif

#define realMsg(fmt, ...)  message_(-1, fmt, ## __VA_ARGS__)

#define errorMsg(fmt, ...)  do { if (!mode_is_silent()) { message_(2, fmt, ## __VA_ARGS__); } } while(0)
#define warnMsg(fmt, ...)   do { if (!mode_is_silent()) { message_(1, fmt, ## __VA_ARGS__); } } while(0)
#define infoMsg(fmt, ...)   do { if (!mode_is_silent()) { message_(0, fmt, ## __VA_ARGS__); } } while(0)

#endif

